
#include "GPumpSettingsData.h"

#define GPUMP_SETTINGS_DATA_GET_PRIVATE(obj)\
  (G_TYPE_INSTANCE_GET_PRIVATE ((obj), GPUMP_TYPE_SETTINGS_DATA,\
                                GPumpSettingsDataPrivate))

G_DEFINE_TYPE (GPumpSettingsData, gpump_settings_data, G_TYPE_OBJECT)


/* Private datastructure */
struct GPumpSettingsDataPrivate {
#if GPUMP_USE_GSETTINGS
  /* org.gego.gpump.ui */
  GSettings * ui_settings;
  /* org.gego.gpump */
  GSettings * gpump_settings;
#else
  GKeyFile * gpump_settings; /* Stores both UI and other settings... */
  GFile * gpump_settings_file;
#endif /* GPUMP_USE_GSETTINGS */
};

/* Properties enumeration */
enum {
  PROP_0,
  
  PROP_UI_X,
  PROP_UI_Y,
  PROP_UI_H,
  PROP_UI_W,
  
  PROP_MAIN_USE_DARK,
  PROP_MAIN_FIRST_RUN,
  
  N_PROPERTIES
};

static GParamSpec * obj_properties[N_PROPERTIES] = { NULL, };



/* * Override the default constructor *****************************************/
GObject * _constructor (GType type,
                        guint n_construct_params,
                        GObjectConstructParam * construct_params);

/* * internal private functions ***********************************************/
void _print_errors (GError ** err_arr);

void gpump_settings_data_set_property (GObject * object,
                                       guint property_id,
                                       const GValue * value,
                                       GParamSpec * pspec);

void gpump_settings_data_get_property (GObject * object,
                                       guint property_id,
                                       GValue * value,
                                       GParamSpec * pspec);

/* * **************************************************************************/
GPumpSettingsData * gpump_settings_data_get_default () {
  GPumpSettingsData * self = g_object_new (GPUMP_TYPE_SETTINGS_DATA,
                                           NULL);
  
  return self;
}

void gpump_settings_data_init (GPumpSettingsData * self) {
  g_print ("Init GPumpSettingsData\n");
  
  self->priv = GPUMP_SETTINGS_DATA_GET_PRIVATE (self);
  
#if GPUMP_USE_GSETTINGS
  /* This will not be compiled, this will have to be implemented later...
   * much later...
   */
  self->ui_settings = g_settings_new ("org.gego.gpump.ui");
  self->gpump_settings = g_settings_new ("org.gego.gpump");
#else
  
  /* Construct the path for the files, 1) the GPump settings 2) the GPump UI 
   * settings.
   *
   * The file names are defined static, so they do not need to be re-created,
   * each and every time to object is re-created.
   */
  self->priv->gpump_settings = g_key_file_new ();
  
  static gchar * gpump_settings_file_uri = NULL;
  static gchar * settings_basedir = NULL;
  
  if (!(gpump_settings_file_uri)) {
    
    g_print ("Constructing path!\n");
    
    gchar * base_config_dir = g_get_user_config_dir ();
    
    settings_basedir = g_strdup_printf ("%s/%s",
                                                base_config_dir,
                                                "gpump");
    g_free (base_config_dir);
    
    gpump_settings_file_uri = g_strdup_printf ("%s/%s",
                                               settings_basedir,
                                               "gpump.conf");
  }
  
  self->priv->gpump_settings_file = g_file_new_for_path (
                                                      gpump_settings_file_uri);
  GError ** err_arr = NULL;
  /* Dose the file exist? */
  if (!g_file_test (gpump_settings_file_uri, G_FILE_TEST_EXISTS )) {
    /* Nope, it did not exist */
    g_print ("Config file does not exist, creating it.\n");
    
    g_mkdir_with_parents (settings_basedir, 0700);
    
    GFileOutputStream * file_stream = g_file_create (
                                                self->priv->gpump_settings_file,
                                                G_FILE_CREATE_PRIVATE,
                                                NULL,
                                                err_arr);
    
    _print_errors (err_arr);
    
    /* Construct a set of default settings */
    
    /* UI settings */
    /* -1 is undefined, these are set when the main window is displayed */
    g_key_file_set_integer (self->priv->gpump_settings, "ui","x", -1);
    g_key_file_set_integer (self->priv->gpump_settings, "ui","y", -1);
    /* 500x500 is the default window size */
    g_key_file_set_integer (self->priv->gpump_settings, "ui","h", 500);
    g_key_file_set_integer (self->priv->gpump_settings, "ui","w", 500);
    /* main settings */
    g_key_file_set_boolean (self->priv->gpump_settings,
                            "main",
                            "use dark",
                            FALSE);
    g_key_file_set_boolean (self->priv->gpump_settings,
                            "main",
                            "first run",
                            TRUE);
    
    g_assert (err_arr == NULL);
    
    gchar * buff = g_key_file_to_data (self->priv->gpump_settings,
                                       NULL,
                                       err_arr);
    
    _print_errors (err_arr);
    
    g_output_stream_write (G_OUTPUT_STREAM (file_stream),
                           buff,
                           strlen (buff),
                           NULL,
                           err_arr);
    
    _print_errors (err_arr);
    
    g_object_unref (file_stream);
  } else {
    /* yes, the file did exit */
    g_key_file_load_from_file (self->priv->gpump_settings,
                               gpump_settings_file_uri,
                               G_KEY_FILE_NONE,
                               err_arr);
    _print_errors (err_arr);
    
  }
#endif
}


void gpump_settings_data_class_init (GPumpSettingsDataClass * klass) {
  GObjectClass * parent = G_OBJECT_CLASS (klass);
  
  parent->constructor = _constructor;
  
  parent->get_property = gpump_settings_data_get_property;
  parent->set_property = gpump_settings_data_set_property;
  
  obj_properties[PROP_UI_X] = g_param_spec_int ("ui-x",
                                                "UI-X",
                                              "The x coordinate of the window",
                                                -1,
                                                INT_MAX,
                                                -1,
                                                G_PARAM_READABLE |
                                                G_PARAM_WRITABLE);
  
  obj_properties[PROP_UI_Y] = g_param_spec_int ("ui-y",
                                                "UI-Y",
                                              "The Y coordinate of the window",
                                                -1,
                                                INT_MAX,
                                                -1,
                                                G_PARAM_READABLE |
                                                G_PARAM_WRITABLE);
  
  obj_properties[PROP_UI_H] = g_param_spec_int ("ui-h",
                                                "UI-h",
                                              "The x coordinate of the window",
                                                1,
                                                INT_MAX,
                                                500,
                                                G_PARAM_READABLE |
                                                G_PARAM_WRITABLE);
  
  obj_properties[PROP_UI_W] = g_param_spec_int ("ui-w",
                                                "UI-w",
                                              "The x coordinate of the window",
                                                1,
                                                INT_MAX,
                                                500,
                                                G_PARAM_READABLE |
                                                G_PARAM_WRITABLE);
  
  obj_properties[PROP_MAIN_USE_DARK] = g_param_spec_boolean ("main-use-dark",
                                                "MAIN-USE-DARK",
                                            "Weather or not to use dark theme.",
                                                FALSE,
                                                G_PARAM_READABLE |
                                                G_PARAM_WRITABLE);
  
  obj_properties[PROP_MAIN_FIRST_RUN] = g_param_spec_boolean ("main-use-dark",
                                                "MAIN-USE-DARK",
                   "Weather or not this is the first time that the app is run.",
                                                TRUE,
                                                G_PARAM_READABLE |
                                                G_PARAM_WRITABLE);
  
  g_object_class_install_properties (parent, N_PROPERTIES,
                                     obj_properties);
  
  g_type_class_add_private (klass, sizeof (GPumpSettingsDataPrivate));
}


void gpump_settings_data_set (GPumpSettingsData * self,
                              gchar * name,
                              GValue * data) {
  g_object_set_property (G_OBJECT (self),
                         name,
                         data);
}


GValue * gpump_settings_data_get (GPumpSettingsData * self,
                                    gchar * name) {
  GValue * data = G_VALUE_INIT;
  g_object_set_property (G_OBJECT (self),
                         name,
                         data);
  return data;
}


gboolean gpump_settings_data_commit (GPumpSettingsData * self,
                                     GError ** err_arr) {
  g_object_ref (G_OBJECT(self));
#if GPUMP_USE_GSETTINGS
  /* Do thothing */
  return TRUE;
#else
  
  g_print ("Committing data to file!\n");
  g_assert ( err_arr == NULL || *err_arr == NULL);
  
  GFileOutputStream * out_stream = g_file_replace (
                                                self->priv->gpump_settings_file,
                                                NULL,
                                                FALSE,
                                                G_FILE_CREATE_PRIVATE,
                                                NULL,
                                                err_arr);
  if (err_arr[0]) {
    g_object_unref (out_stream);
    return FALSE;
  }
  
  gsize * len = NULL;
  gchar * buff = g_key_file_to_data (self->priv->gpump_settings, len, err_arr);
  
  if (err_arr) {
   g_object_unref (out_stream);
   g_free (len);
   g_free (buff);
   return FALSE;
  }
  
  g_output_stream_write (G_OUTPUT_STREAM (out_stream),
                         buff,
                         (gsize) len,
                         NULL,
                         err_arr);
  
  if (err_arr) {
    g_object_unref (out_stream);
    g_free (len);
    g_free (buff);
    return FALSE;
  }
  
  g_object_unref (out_stream);
  g_free (len);
  g_free (buff);
  
  return TRUE;
#endif
  g_object_unref (G_OBJECT (self));
}

/* * **************************************************************************
 * * Implementation of the constructor ****************************************
 *
 * See: blogs.gnome.org/xclaesse/2010/02/11/how-to-make-a-gobject-singleton/
 * for more information.
 */
GObject * _constructor (GType type,
                        guint n_construct_params,
                        GObjectConstructParam * construct_params) {
  static GObject * self = NULL;
  static GMutex mutex;
  
  g_mutex_lock (&mutex);
  if (self == NULL) {
    self = G_OBJECT_CLASS (gpump_settings_data_parent_class)->constructor (
      type, n_construct_params, construct_params);
    g_object_add_weak_pointer (self, (gpointer) &self);
    g_mutex_unlock (&mutex);
    return self;
  }
  g_mutex_unlock (&mutex);
  return g_object_ref (self);
}

/* * Internal private functions ***********************************************/
void _print_errors (GError ** err_arr) {
  if (err_arr) {
    for (int i = 0; err_arr[i] != NULL; i++) {
       g_print ("GPumpSettingsData:\n An %s error occured:\n\t%s\n\n",
                 g_quark_to_string (err_arr[i]->domain),
                 err_arr[i]->message);
    }
    abort ();
  }
}

/* * **************************************************************************
 * * Getting and setting properties *******************************************
 * For more information see:
 * https://developer.gnome.org/gobject/stable/gobject-properties.html
 */

/* * declarations of setting functions */
void _set_ui_x (GPumpSettingsData * self,
                GValue * value);
void _set_ui_y (GPumpSettingsData * self,
                GValue * value);
void _set_ui_h (GPumpSettingsData * self,
                GValue * value);
void _set_ui_w (GPumpSettingsData * self,
                GValue * value);

void _set_main_use_dark (GPumpSettingsData * self,
                         GValue * value);
void _set_main_first_run (GPumpSettingsData * self,
                          GValue * value);


/* * declarations of getting functions */
void _get_ui_x (GPumpSettingsData * self,
                GValue * value);
void _get_ui_y (GPumpSettingsData * self,
                GValue * value);
void _get_ui_h (GPumpSettingsData * self,
                GValue * value);
void _get_ui_w (GPumpSettingsData * self,
                GValue * value);

void _get_main_use_dark (GPumpSettingsData * self,
                         GValue * value);
void _get_main_first_run (GPumpSettingsData * self,
                          GValue * value);



/* * Setting props */
void gpump_settings_data_set_property (GObject * object,
                                       guint property_id,
                                       const GValue * value,
                                       GParamSpec * pspec) {
  g_object_ref (object);
  
  GPumpSettingsData * self = GPUMP_SETTINGS_DATA (object);
  
  switch (property_id) {
    
    /* UI Props */
    case PROP_UI_X:
      _set_ui_x (self, value);
      break;
    case PROP_UI_Y:
      _set_ui_x (self, value);
      break;
    case PROP_UI_H:
      _set_ui_x (self, value);
      break;
    case PROP_UI_W:
      _set_ui_x (self, value);
      break;
    
    /* GPump generic props */
    case PROP_MAIN_USE_DARK:
      _set_main_use_dark (self, value);
      break;
    case PROP_MAIN_FIRST_RUN:
      _set_main_first_run (self, value);
      break;
    
    /* Default */
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
      break;
    
  }
  g_object_unref (object);
}

/* * Getting */
void gpump_settings_data_get_property (GObject * object,
                                       guint property_id,
                                       GValue * value,
                                       GParamSpec * pspec) {
  g_object_ref (object);
  
  GPumpSettingsData * self = GPUMP_SETTINGS_DATA (object);
  
  switch (property_id) {
    
    /* UI Props */
    case PROP_UI_X:
      _get_ui_x (self, value);
      break;
    case PROP_UI_Y:
      _get_ui_x (self, value);
      break;
    case PROP_UI_H:
      _get_ui_x (self, value);
      break;
    case PROP_UI_W:
      _get_ui_x (self, value);
      break;
    
    /* GPump generic props */
    case PROP_MAIN_USE_DARK:
      _get_main_use_dark (self, value);
      break;
    case PROP_MAIN_FIRST_RUN:
      _get_main_first_run (self, value);
      break;
    
    /* Default */
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
      break;
    
  }
  g_object_unref (object);
}


/* * implimentations of setting functions *************************************/
void _set_ui_x (GPumpSettingsData * self,
                GValue * value) {
#if GPUMP_USE_GSETTINGS
  /* TODO */
#else
  g_key_file_set_integer (self->priv->gpump_settings,
                          "ui",
                          "x",
                          g_value_get_int (value));
#endif
}

void _set_ui_y (GPumpSettingsData * self,
                GValue * value) {
#if GPUMP_USE_GSETTINGS
  /* TODO */
#else
  g_key_file_set_integer (self->priv->gpump_settings,
                          "ui",
                          "y",
                          g_value_get_int (value));
#endif
}

void _set_ui_h (GPumpSettingsData * self,
                GValue * value) {
#if GPUMP_USE_GSETTINGS
  /* TODO */
#else
  g_key_file_set_integer (self->priv->gpump_settings,
                          "ui",
                          "h",
                          g_value_get_int (value));
#endif
}

void _set_ui_w (GPumpSettingsData * self,
                GValue * value) {
#if GPUMP_USE_GSETTINGS
  /* TODO */
#else
  g_key_file_set_integer (self->priv->gpump_settings,
                          "ui",
                          "w",
                          g_value_get_int (value));
#endif
}

void _set_main_use_dark (GPumpSettingsData * self,
                         GValue * value) {
#if GPUMP_USE_GSETTINGS
  /* TODO */
#else
  g_key_file_set_integer (self->priv->gpump_settings,
                          "main",
                          "use dark",
                          g_value_get_boolean (value));
#endif
}

void _set_main_first_run (GPumpSettingsData * self,
                          GValue * value)  {
#if GPUMP_USE_GSETTINGS
  /* TODO */
#else
  g_key_file_set_integer (self->priv->gpump_settings,
                          "main",
                          "first run",
                          g_value_get_boolean (value));
#endif
}


/* * implimentations of getting functions *************************************/
void _get_ui_x (GPumpSettingsData * self,
                GValue * value) {
#if GPUMP_USE_GSETTINGS
  /* TODO */
#else
  GError ** err_arr = NULL;
  g_value_set_int (value,
                   g_key_file_get_integer (self->priv->gpump_settings,
                                           "ui",
                                           "x",
                                           err_arr));
  _print_errors (err_arr);
  g_clear_error (err_arr);
#endif
}

void _get_ui_y (GPumpSettingsData * self,
                GValue * value) {
#if GPUMP_USE_GSETTINGS
  /* TODO */
#else
  GError ** err_arr = NULL;
  g_value_set_int (value,
                   g_key_file_get_integer (self->priv->gpump_settings,
                                           "ui",
                                           "y",
                                           err_arr));
  _print_errors (err_arr);
  g_clear_error (err_arr);
#endif
}

void _get_ui_h (GPumpSettingsData * self,
                GValue * value) {
#if GPUMP_USE_GSETTINGS
  /* TODO */
#else
  GError ** err_arr = NULL;
  g_value_set_int (value,
                   g_key_file_get_integer (self->priv->gpump_settings,
                                           "ui",
                                           "h",
                                           err_arr));
  _print_errors (err_arr);
  g_clear_error (err_arr);
#endif
}

void _get_ui_w (GPumpSettingsData * self,
                GValue * value) {
#if GPUMP_USE_GSETTINGS
  /* TODO */
#else
  GError ** err_arr = NULL;
  g_value_set_int (value,
                   g_key_file_get_integer (self->priv->gpump_settings,
                                           "ui",
                                           "w",
                                           err_arr));
  _print_errors (err_arr);
  g_clear_error (err_arr);
#endif
}

void _get_main_use_dark (GPumpSettingsData * self,
                         GValue * value) {
#if GPUMP_USE_GSETTINGS
  /* TODO */
#else
  GError ** err_arr = NULL;
  g_value_set_int (value,
                   g_key_file_get_integer (self->priv->gpump_settings,
                                           "main",
                                           "use dark",
                                           err_arr));
  _print_errors (err_arr);
  g_clear_error (err_arr);
#endif
}

void _get_main_first_run (GPumpSettingsData * self,
                          GValue * value)  {
#if GPUMP_USE_GSETTINGS
  /* TODO */
#else
  GError ** err_arr = NULL;
  g_value_set_int (value,
                   g_key_file_get_integer (self->priv->gpump_settings,
                                           "main",
                                           "first run",
                                           err_arr));
  _print_errors (err_arr);
  g_clear_error (err_arr);
#endif
}

